@@ -1,6 +1,6 @@
 package App::ansicolumn;
 
-our $VERSION = "1.56";
+our $VERSION = "1.57";
 
 use v5.14;
 use warnings;
@@ -30,11 +30,12 @@ sub ansicolumn {
 }
 
 my %DEFAULT_COLORMAP = (
-    BORDER => '',
-    TEXT   => '',
+    BORDER       => '',
+    BORDER_LABEL => '',
+    TEXT          => '',
 );
 
-use Getopt::EX::Hashed 1.05; {
+use Getopt::EX::Hashed 1.07; {
 
     Getopt::EX::Hashed->configure( DEFAULT => [ is => 'rw' ] );
 
@@ -88,6 +89,8 @@ use Getopt::EX::Hashed 1.05; {
     has border              => ' :s      ' ; has B => '' , action => sub { $_->border = '' } ;
     has no_border           => '         ' , action => sub { $_->border = 'none' } ;
     has border_style        => ' =s   bs ' , default => 'box' ;
+    has label               => ' =s%     ' , default => {} ;
+    has page_label          => ' =s%     ' , default => {} ;
     has white_space         => ' !       ' , default => 2 ;
     has isolation           => ' !       ' , default => 2 ;
     has fillup              => ' :s      ' ; has F => '' , action => sub { $_->fillup = '' } ;
@@ -113,6 +116,8 @@ use Getopt::EX::Hashed 1.05; {
     has panes         => ;
     has border_height => ;
     has current_page  => ;
+    has _bl           => ;
+    has _pbl          => ;
 
     Getopt::EX::Hashed->configure( DEFAULT => [] );
 
@@ -343,6 +348,7 @@ sub parallel_out {
 	}
     }
     $obj->set_contents($_) for @files;
+    my $column_out = $obj->_column_out;
     while (@files) {
 	my @rows = splice @files, 0, $obj->pane;
 	my $max_length = max map { int @{$_->{data}} } @rows;
@@ -359,8 +365,7 @@ sub parallel_out {
 		ansi_sprintf $obj->filename_format, $_->{name};
 	    } @rows;
 	}
-	$obj->column_out(
-	    { span => \@span },
+	$column_out->({ span => \@span, names => [ map $_->{name}, @rows ] },
 	    map {
 		my $data = $_->{data};
 		my $length = @$data;
@@ -383,7 +388,7 @@ sub nup_out {
 	    ->set_contents($file)
 	    ->set_vertical($data)
 	    ->set_layout($data)
-	    ->page_out(@$data);
+	    ->page_out($file->{name}, @$data);
     }
     return $obj;
 }
@@ -494,6 +499,8 @@ sub set_vertical {
 
 sub page_out {
     my $obj = shift;
+    my $name = shift;
+    my $column_out = $obj->_column_out;
     for ($obj->current_page = 0; @_; $obj->current_page++) {
 	my @columns = grep { @$_ } do {
 	    if ($obj->fillrows) {
@@ -502,7 +509,7 @@ sub page_out {
 		map { [ splice @_, 0, $obj->effective_height ] } 1 .. $obj->panes;
 	    }
 	};
-	$obj->column_out(@columns);
+	$column_out->({ names => [($name) x scalar @columns] }, @columns);
     }
     return $obj;
 }
@@ -512,37 +519,165 @@ sub color_border {
     $obj->color('BORDER', $obj->get_border(@_));
 }
 
-sub column_out {
+sub _parse_labels {
+    my $hash = shift;
+    return 0 unless %$hash;
+    my %result;
+    while (my($key, $s) = each %$hash) {
+	next unless defined $s and $s ne '';
+	my $offset;
+	if ($s =~ s/@(\d+)$//) {
+	    $offset = $1 + 0;
+	}
+	$s =~ s/%n/%1\$d/g;
+	$s =~ s/%p/%2\$d/g;
+	$s =~ s/%f/%3\$s/g;
+	$s =~ s/%F/%4\$s/g;
+	$result{lc $key} = [ $offset, $s ];
+    }
+    %result ? \%result : 0;
+}
+
+# Overlay labels onto a border line string.
+# $fs/$fe: fill region start/end.
+# $ml/$mr: border margin accessible via @0.
+# $n/$p: pane number / page number for sprintf expansion.
+# $file: filename (path as given) - used as %3$s (basename) and %4$s (path as given).
+sub _overlay_labels {
+    my($obj, $str, $fs, $fe, $ml, $mr, $n, $p, $file, $left, $center, $right) = @_;
+    use File::Basename qw(basename);
+    use Text::ANSI::Fold::Util qw(ansi_substr);
+    my $base = defined $file ? basename($file) : '';
+    for my $item (
+	[ $left,   sub { $fs - $ml + ($_[0] // $ml) } ],
+	[ $right,  sub { $fe + $mr - ($_[0] // $mr) - $_[1] } ],
+	[ $center, sub { $fs + int(($fe - $fs - $_[1]) / 2) } ],
+    ) {
+	my($spec, $pos_fn) = @$item;
+	next unless $spec;
+	my $fmt = $spec->[1];
+	next unless defined $fmt && $fmt ne '';
+	my $text = $fmt =~ /%/ ? sprintf($fmt, $n, $p, $base, $file // '') : $fmt;
+	my $lw = ansi_width($text);
+	my $label = $obj->color('BORDER_LABEL', $text);
+	$str = ansi_substr($str, $pos_fn->($spec->[0], $lw), $lw, $label);
+    }
+    $str;
+}
+
+sub _column_out {
     my $obj = shift;
-    my $opt = ref $_[0] eq 'HASH' ? shift : {};
 
-    # span list is given in parallel view mode
-    my @span = $opt->{span} ? @{$opt->{span}} : (($obj->{span}) x @_);
-    @span == @_ or die;
-
-    # insert top/bottom border
     my %bd = map { $_ => $obj->get_border($_) } qw(top bottom);
-    if ($bd{top} or $bd{bottom}) {
-	while (my($i, $e) = each @_) {
-	    unshift @$e, $obj->color('BORDER', $bd{top} x $span[$i]) if $bd{top};
-	    push @$e, $obj->color('BORDER', $bd{bottom} x $span[$i]) if $bd{bottom};
+
+    my $bl  = $obj->{_bl}  //= _parse_labels($obj->label);
+    my $pbl = $obj->{_pbl} //= _parse_labels($obj->page_label);
+
+    # pre-build pane label closures per pane index
+    my(@pane_top, @pane_bottom, @page_top, @page_bottom);
+    my($bw_l, $bw_c, $bw_r);
+    if ($bl || $pbl) {
+	$bw_l = ansi_width($obj->get_border('left',   0));
+	$bw_c = ansi_width($obj->get_border('center', 0));
+	$bw_r = ansi_width($obj->get_border('right',  0));
+	my $span = $obj->{span};
+	for my $set ([ \@pane_top, \@page_top, qw(nw n ne) ],
+		     [ \@pane_bottom, \@page_bottom, qw(sw s se) ]) {
+	    my($pane_list, $page_list, @keys) = @$set;
+	    if ($bl) {
+		my @g = map { $bl->{$_} } @keys;
+		if (grep { $_ } @g) {
+		    my $fp = $bw_l;
+		    for my $i (0 .. $obj->panes - 1) {
+			my($fs, $fe) = ($fp, $fp + $span);
+			push @$pane_list, sub {
+			    $obj->_overlay_labels($_[0], $fs, $fe, $bw_l, $bw_r,
+				$_[1] + $i + 1, $_[2], $_[3], @g);
+			};
+			$fp += $span + $bw_c;
+		    }
+		}
+	    }
+	    if ($pbl) {
+		my @g = map { $pbl->{$_} } @keys;
+		if (grep { $_ } @g) {
+		    push @$page_list, sub {
+			my $w = ansi_width($_[0]);
+			$obj->_overlay_labels($_[0], 1, $w - 1, 1, 1,
+			    $_[1] + 1, $_[2], $_[3], @g);
+		    };
+		}
+	    }
 	}
     }
 
-    my $max = max map $#{$_}, @_;
-    for my $i (0 .. $max) {
-	my $pos = $i == 0 ? 0 : $i == $max ? 2 : 1;
-	my @span = @span;
-	my @panes = map {
-	    @$_ ? ansi_sprintf("%-*s", shift @span, shift @$_) : ();
-	} @_;
-	print      $obj->color_border('left',   $pos, $obj->current_page);
-	print join $obj->color_border('center', $pos, $obj->current_page),
-	    map { $obj->color('TEXT', $_) } @panes;
-	print      $obj->color_border('right',  $pos, $obj->current_page);
-	print      "\n";
+    my $has_labels = @pane_top || @pane_bottom || @page_top || @page_bottom;
+
+    # pre-build border strings for 3 positions (top=0, middle=1, bottom=2)
+    my %border;
+    for my $pos (0, 1, 2) {
+	$border{$pos} = {
+	    left   => $obj->color_border('left',   $pos),
+	    center => $obj->color_border('center', $pos),
+	    right  => $obj->color_border('right',  $pos),
+	};
     }
-    return $obj;
+
+    my $print_border = sub {
+	my($side, $pos, $span, $npanes, $n0, $p, $names,
+	   $pane_labels, $page_labels) = @_;
+	return unless $bd{$side};
+	my $bd = $border{$pos};
+	my $line = join '',
+	    $bd->{left},
+	    join($bd->{center},
+		 map { $obj->color('BORDER', $bd{$side} x $_) } @$span),
+	    $bd->{right};
+	if (@$pane_labels) {
+	    my $i = 0;
+	    for (@{$pane_labels}[0 .. $npanes - 1]) {
+		$line = $_->($line, $n0, $p, $names->[$i++] // '');
+	    }
+	}
+	$line = $_->($line, $n0, $p, $names->[0] // '') for @$page_labels;
+	print $line, "\n";
+    };
+
+    sub {
+	my $opt = ref $_[0] eq 'HASH' ? shift : {};
+	my @span = $opt->{span} ? @{$opt->{span}} : (($obj->{span}) x @_);
+	my $page = $obj->current_page // 0;
+	my $npanes = scalar @_;
+
+	my($n0, $p) = $has_labels
+	    ? ($page * ($obj->panes // 1), $page + 1) : ();
+	my $names = $opt->{names} // [];
+
+	$print_border->('top', 0, \@span, $npanes, $n0, $p, $names,
+			 \@pane_top, \@page_top);
+
+	# content rows
+	my $max = max map $#{$_}, @_;
+	for my $i (0 .. $max) {
+	    my $pos = !$bd{top} && $i == 0 ? 0
+		    : !$bd{bottom} && $i == $max ? 2
+		    : 1;
+	    my $bd = $border{$pos};
+	    my @span = @span;
+	    my @panes = map {
+		@$_ ? ansi_sprintf("%-*s", shift @span, shift @$_) : ();
+	    } @_;
+	    print      $bd->{left};
+	    print join $bd->{center},
+		map { $obj->color('TEXT', $_) } @panes;
+	    print      $bd->{right};
+	    print      "\n";
+	}
+
+	$print_border->('bottom', 2, \@span, $npanes, $n0, $p, $names,
+			 \@pane_bottom, \@page_bottom);
+	$obj;
+    };
 }
 
 sub _numbers {
